VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
END
Attribute VB_Name = "DebugSynchClass"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = True
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = False
''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' John Robbins
' Microsoft Systems Journal - August, 1997
''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' FILE          :   DebugSynchClass.cls
' DESCRIPTION   :
'  The class that handles the chores of encapsulating all of the
'  synchronization object that are used to coordinate the debugger
'  thread and the user interface.
'  When looking this over, you might wonder why there is no
'  synchronization handling to indicate that the debug thread is fully
'  shut down.  While it could be done, it would be redundant, because
'  the UI should be triggered off of the debug thread handle.  That's
'  the only way to ensure that a thread is really done processing.
'  Each thread MUST create their own instances of this class.
'
''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
Option Explicit

''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Class private constants
''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' The event that signals when I had a good debuggee process startup.
'  The PID is appended to it.
Private Const k_GOODSTARTUPSTRING As String = "VBDEBUG_GOODSTARTUP"
' The event that signals when I had a failure of the debuggee process
'  startup.  The PID is appended to it.
Private Const k_BADSTARTUPSTRING As String = "VBDEBUG_BADSTARTUP"

' The array indexes for the startup events.
Private Const k_GOODSTARTID As Long = 0
Private Const k_BADSTARTID As Long = 1

' The core synchronization objects:  the debugging event, quit, pause,
'  and resume events.  These have the DEBUGGEE process ID appended to
'  them to make sure that they are unique.
Private Const k_QUITSTRING As String = "VBDEBUG_QUIT"
Private Const k_PAUSESTRING As String = "VBDEBUG_PAUSE"
Private Const k_RESUMESTRING As String = "VBDEBUG_RESUME"
Private Const k_DEBUGSTRING As String = "VBDEBUG_DEBUG"

' The IDs for which elements in the event array belong to which.  Too
'  bad VB won't let you expose constants.
' The quit event is first because the WaitForMultipleObjects function
'  will return on it first when it is signaled before looking at the
'  regular debug events.
Private Const k_QUITEVENT As Long = 0
Private Const k_PAUSEEVENT As Long = 1
Private Const k_RESUMEEVENT As Long = 2
Private Const k_DEBUGACTIVEEVENT As Long = 3

''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Class private variables
''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' The process ID of the debugger.
Private m_hControlPID As Long
' The process ID for the debuggee.
Private m_hDebuggeePID As Long
' The core events I wait on, the quit, pause, resume, and debug.
Private m_hDebugEvents(4) As Long
' The start up events that can be signaled.
Private m_hStartupEvents(2) As Long

''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' FUNCTION        : Class_Initialize
' DISCUSSION      :
'  The initialization for the class.
' PARAMETERS      :
'  None.
''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
Private Sub Class_Initialize()
   m_hDebugEvents(0) = 0
   m_hDebugEvents(1) = 0
   m_hDebugEvents(2) = 0
   m_hDebugEvents(3) = 0
   m_hDebuggeePID = 0
   m_hStartupEvents(0) = 0
   m_hStartupEvents(1) = 0
   m_hControlPID = GetCurrentProcessId()
End Sub

''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' FUNCTION        : Class_Terminate
' DISCUSSION      :
'  The termination for the class.
' PARAMETERS      :
'  None.
''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
Private Sub Class_Terminate()

   ' If the handles have not been destroyed, do it now.
   If ((0 <> m_hDebugEvents(k_QUITEVENT)) And _
      (0 <> m_hDebugEvents(k_DEBUGACTIVEEVENT))) Then
      DeleteSynchObjects
   End If

End Sub

''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
'           Properties and methods only for the debug thread!
''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''

''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' FUNCTION        :  dwUniqueID
' DISCUSSION      :
'  The property set for the unique ID for the core synch events.
' PARAMETERS      :
'  dwID - The value.
''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
Public Property Let dwUniqueID(dwID As Variant)
   m_hDebuggeePID = dwID
End Property

''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' FUNCTION        :  SignalGoodStartup
' DISCUSSION      :
'  Signals that the debuggee started up correctly.  This is only called
'  from the debug thread.  The UI is supposed to be waiting in the
'  WaitForStartup method.
' PARAMETERS      :
'  None.
''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
Public Sub SignalGoodStartup()

   SignalManualResetEvent (k_GOODSTARTUPSTRING + Hex$(m_hControlPID))

End Sub

''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' FUNCTION        :  SignalBadStartup
' DISCUSSION      :
'  Signals that the debuggee failed to start correctly.  This is only
'  called from the debug thread.  The UI is supposed to be waiting in
'  the WaitForStartup method.
' PARAMETERS      :
'  None.
''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
Public Sub SignalBadStartup()

   SignalManualResetEvent (k_BADSTARTUPSTRING + Hex$(m_hControlPID))

End Sub

''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' FUNCTION        :  CreateSynchObjects
' DISCUSSION      :
'  The function that creates the synchronization objects that are used
'  to coordinate everything between the debug thread and the UI.
' PARAMETERS      :
'  None.
' RETURNS         :
'  True  - Everything was created correctly.
'  False - There was a problem.
''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
Public Function CreateSynchObjects() As Boolean

   Dim bRet As Long

   ' I assume that the return value will be good.
   CreateSynchObjects = True

#If DEBUGBUILD Then
   If (0 = m_hDebuggeePID) Then
      MsgBox ("CreateSynchObjects called before the unique id set!")
   End If
#End If

   If (0 = m_hDebuggeePID) Then
      CreateSynchObjects = False
      Exit Function
   End If

   ' Create the quit event.  It is a manual reset event that is not
   '  signaled
   m_hDebugEvents(k_QUITEVENT) = CreateEvent(0, _
                                             1, _
                                             0, _
                                             k_QUITSTRING + _
                                                   Hex$(m_hDebuggeePID))

#If DEBUGBUILD Then
   If (0 = m_hDebugEvents(k_QUITEVENT)) Then
      MsgBox ("CreateSynchObjects unable to create Quit event!")
   End If
#End If

   If (0 = m_hDebugEvents(k_QUITEVENT)) Then
      CreateSynchObjects = False
      Exit Function
   End If

   ' Create the pause event.  It is a manual reset event that is not
   '  signaled
   m_hDebugEvents(k_PAUSEEVENT) = CreateEvent(0, _
                                              1, _
                                              0, _
                                              k_PAUSESTRING + _
                                                   Hex$(m_hDebuggeePID))

#If DEBUGBUILD Then
   If (0 = m_hDebugEvents(k_PAUSEEVENT)) Then
      MsgBox ("CreateSynchObjects unable to create Pause event!")
   End If
#End If

   If (0 = m_hDebugEvents(k_PAUSEEVENT)) Then
      bRet = CloseHandle(m_hDebugEvents(k_QUITEVENT))
      CreateSynchObjects = False
      Exit Function
   End If

   ' Create the resume event.  It is a manual reset event that is not
   '  signaled
   m_hDebugEvents(k_RESUMEEVENT) = CreateEvent(0, _
                                               1, _
                                               0, _
                                               k_RESUMESTRING + _
                                                   Hex$(m_hDebuggeePID))

#If DEBUGBUILD Then
   If (0 = m_hDebugEvents(k_RESUMEEVENT)) Then
      MsgBox ("CreateSynchObjects unable to create Resume event!")
   End If
#End If

   If (0 = m_hDebugEvents(k_RESUMEEVENT)) Then
      bRet = CloseHandle(m_hDebugEvents(k_QUITEVENT))
      bRet = CloseHandle(m_hDebugEvents(k_PAUSEEVENT))
      CreateSynchObjects = False
      Exit Function
   End If

   ' Create the debug event.  It is a manual reset event that is
   '  signaled.
   m_hDebugEvents(k_DEBUGACTIVEEVENT) = CreateEvent(0, _
                                                    1, _
                                                    1, _
                                                    k_DEBUGSTRING + _
                                                   Hex$(m_hDebuggeePID))
#If DEBUGBUILD Then
   If (0 = m_hDebugEvents(k_DEBUGACTIVEEVENT)) Then
      MsgBox ("CreateSynchObjects unable to create Debug event!")
   End If
#End If

   If (0 = m_hDebugEvents(k_DEBUGACTIVEEVENT)) Then
      bRet = CloseHandle(m_hDebugEvents(k_QUITEVENT))
      bRet = CloseHandle(m_hDebugEvents(k_PAUSEEVENT))
      bRet = CloseHandle(m_hDebugEvents(k_RESUMEEVENT))
      CreateSynchObjects = False
      Exit Function
   End If

End Function

''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' FUNCTION        :  DeleteSynchObjects
' DISCUSSION      :
'  Deletes the synchronization objects created with CreateSynchObjects.
' PARAMETERS      :
'  None.
''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
Public Sub DeleteSynchObjects()

   Dim bRet As Long

#If DEBUGBUILD Then
   If ((0 = m_hDebugEvents(k_DEBUGACTIVEEVENT)) Or _
      ((0 = m_hDebugEvents(k_QUITEVENT)))) Then
      MsgBox ("DeleteSynchObjects called before CreateSynchObjects!")
   End If
#End If

   bRet = CloseHandle(m_hDebugEvents(k_DEBUGACTIVEEVENT))
   m_hDebugEvents(k_DEBUGACTIVEEVENT) = 0
#If DEBUGBUILD Then
   If (0 = bRet) Then
      MsgBox ("DeleteSynchObjects failed closing debug event handle")
   End If
#End If

   bRet = CloseHandle(m_hDebugEvents(k_QUITEVENT))
   m_hDebugEvents(k_QUITEVENT) = 0
#If DEBUGBUILD Then
   If (0 = bRet) Then
      MsgBox ("DeleteSynchObjects failed closing quit event handle")
   End If
#End If

   bRet = CloseHandle(m_hDebugEvents(k_PAUSEEVENT))
   m_hDebugEvents(k_PAUSEEVENT) = 0
#If DEBUGBUILD Then
   If (0 = bRet) Then
      MsgBox ("DeleteSynchObjects failed closing pause event handle")
   End If
#End If

   bRet = CloseHandle(m_hDebugEvents(k_RESUMEEVENT))
   m_hDebugEvents(k_RESUMEEVENT) = 0
#If DEBUGBUILD Then
   If (0 = bRet) Then
      MsgBox ("DeleteSynchObjects failed closing resume event handle")
   End If
#End If

End Sub

''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' FUNCTION        :  WaitForSynchObject
' DISCUSSION      :
'  This is the function that the debug thread will call to figure out
'  what it should be doing.
' PARAMETERS      :
'  The maximum time, in milliseconds to wait.  The SDK define INFINITE
'  is permitted.
' RETURNS         :
'  0         - The quit event was signaled.
'  1         - The pause event was signaled.
'  2         - The resume event was signaled.
'  3         - The debug event was signaled.
'  Otherwise - See the return values for WaitForMultipleObjects
''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
Public Function WaitForSynchObject(dwWaitTime As Long) As Long

#If DEBUGBUILD Then
   If ((0 = m_hDebugEvents(k_DEBUGACTIVEEVENT)) Or _
      ((0 = m_hDebugEvents(k_QUITEVENT)))) Then
      MsgBox ("DeleteSynchObjects called before CreateSynchObjects!")
   End If
#End If

   Dim dwRet As Long
   dwRet = WaitForMultipleObjects(4, _
                                  m_hDebugEvents(0), _
                                  0, _
                                  dwWaitTime)
   ' Handle the resets on Pause and Resume.
   Dim bTemp As Long
   If (k_PAUSEEVENT = dwRet) Then
      bTemp = ResetEvent(m_hDebugEvents(k_DEBUGACTIVEEVENT))
      ' Put the pause back into its normal state.
      bTemp = ResetEvent(m_hDebugEvents(k_PAUSEEVENT))
   ElseIf (k_RESUMEEVENT = dwRet) Then
      bTemp = SetEvent(m_hDebugEvents(k_DEBUGACTIVEEVENT))
      bTemp = ResetEvent(m_hDebugEvents(k_RESUMEEVENT))
   End If

   WaitForSynchObject = dwRet
End Function

''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
'           Properties and methods only for the UI thread!
''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''

''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' FUNCTION        :  PrepareWaitForStartup
' DISCUSSION      :
'  The UI MUST call this function before it can call WaitForStartup.
'  It must also be called before spawning the debug thread as well.
'  As I found out on those really, really fast DEC ALPHAs, the debugger
'  thread can fail the CreateProcess and call SignalBadStartup
'  before the UI can call WaitForStartup, which, you guessed it, causes
'  massive deadlock.  By requiring the UI to call this function first,
'  I make sure that the event is ready.  400Mhz really helps show
'  synchronization problems!
' PARAMETERS      :
'  None.
''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
Public Sub PrepareWaitForStartup()

   ' Create the events
   m_hStartupEvents(k_GOODSTARTID) = CreateEvent(0, _
                                                 1, _
                                                 0, _
                                                 k_GOODSTARTUPSTRING + _
                                                    Hex$(m_hControlPID))

   m_hStartupEvents(k_BADSTARTID) = CreateEvent(0, _
                                                1, _
                                                0, _
                                                k_BADSTARTUPSTRING + _
                                                    Hex$(m_hControlPID))

#If DEBUGBUILD Then
   If ((0 = m_hStartupEvents(k_BADSTARTID)) Or _
       (0 = m_hStartupEvents(k_GOODSTARTID))) Then
      MsgBox ("PrepareWaitForStartup failed to create events")
   End If
#End If

End Sub

''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' FUNCTION        :  WaitForStartup
' DISCUSSION      :
'  The UI thread calls this after it has created the thread so that it
'  can then check if life got cranked up.  The UI must call
'  PrepareWaitForStartup FIRST!
' PARAMETERS      :
'  None.
' RETURNS         :
'  TRUE  - The debuggee was started correctly.
'  FALSE - The debuggee was not started correctly.
''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
Public Function WaitForStartup() As Boolean

#If DEBUGBUILD Then
   If ((0 = m_hStartupEvents(k_BADSTARTID)) Or _
       (0 = m_hStartupEvents(k_GOODSTARTID))) Then
      MsgBox ("WaitForStartup can only be called after " + _
               "PrepareWaitForStartup!")
   End If
#End If

   Dim dwRet As Long
   Dim bRet As Long

   ' Wait for one of the startup events.
   dwRet = WaitForMultipleObjects(2, _
                                  m_hStartupEvents(0), _
                                  0, _
                                  INFINITE)


   ' Close the handles.
   bRet = CloseHandle(m_hStartupEvents(k_BADSTARTID))

#If DEBUGBUILD Then
   If (0 = bRet) Then
      MsgBox ("WaitForStartup CloseHandle(k_BADSTARTID) failed!")
   End If
#End If

   bRet = CloseHandle(m_hStartupEvents(k_GOODSTARTID))

#If DEBUGBUILD Then
   If (0 = bRet) Then
      MsgBox ("WaitForStartup CloseHandle(k_GOODSTARTID) failed!")
   End If
#End If

   If (k_GOODSTARTID = dwRet) Then
      WaitForStartup = True
   Else
      WaitForStartup = False
   End If

End Function

''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' FUNCTION        :  PauseDebugThread
' DISCUSSION      :
'  Pauses the debug thread.  Only the UI thread should call this!
' PARAMETERS      :
'  None.
''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
Public Sub PauseDebugThread()

#If DEBUGBUILD Then
   If ((0 = m_hDebugEvents(k_DEBUGACTIVEEVENT)) Or _
      ((0 = m_hDebugEvents(k_QUITEVENT)))) Then
      MsgBox ("PauseDebugThread called before CreateSynchObjects!")
   End If
#End If

   Dim bRet As Long
   bRet = SetEvent(m_hDebugEvents(k_PAUSEEVENT))

#If DEBUGBUILD Then
   If (0 = bRet) Then
      MsgBox ("PauseDebugThread ResetEvent failed!")
   End If
#End If

End Sub

''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' FUNCTION        :  ResumeDebugThread
' DISCUSSION      :
'  Resumes the debug thread.  Only the UI thread should call this!
' PARAMETERS      :
'  None.
''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
Public Sub ResumeDebugThread()

#If DEBUGBUILD Then
   If ((0 = m_hDebugEvents(k_DEBUGACTIVEEVENT)) Or _
      ((0 = m_hDebugEvents(k_QUITEVENT)))) Then
      MsgBox ("ResumeDebugThread called before CreateSynchObjects!")
   End If
#End If

   Dim bRet As Long
   bRet = SetEvent(m_hDebugEvents(k_RESUMEEVENT))

#If DEBUGBUILD Then
   If (0 = bRet) Then
      MsgBox ("ResumeDebugThread SetEvent failed!")
   End If
#End If

End Sub

''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' FUNCTION        :  QuitDebugThread
' DISCUSSION      :
'  Quits the debug thread.  Only the UI thread should call this!
' PARAMETERS      :
'  None.
''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
Public Sub QuitDebugThread()

#If DEBUGBUILD Then
   If ((0 = m_hDebugEvents(k_DEBUGACTIVEEVENT)) Or _
      ((0 = m_hDebugEvents(k_QUITEVENT)))) Then
      MsgBox ("QuitDebugThread called before CreateSynchObjects!")
   End If
#End If

   ' All I have to do is set the debug event to the signaled state.
   Dim bRet As Long
   bRet = SetEvent(m_hDebugEvents(k_QUITEVENT))

#If DEBUGBUILD Then
   If (0 = bRet) Then
      MsgBox ("QuitDebugThread SetEvent failed!")
   End If
#End If

End Sub

''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
'                     Class Private Helper Routines
''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''

''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' FUNCTION        :  SignalManualResetEvent
' DISCUSSION      :
'  A helper function for SignalGoodStartup and SignalBadStartup.  Simply
'  creates and signals the event with the passed in string.
' PARAMETERS      :
'  None.
''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
Private Sub SignalManualResetEvent(szStr As String)

   Dim hEvent As Long
   Dim bRet As Long

   ' Create the event.
   hEvent = CreateEvent(0, _
                        1, _
                        1, _
                        szStr)
#If DEBUGBUILD Then
   If (0 = hEvent) Then
      MsgBox ("SignalManualResetEvent failed to create event")
   End If
#End If

   If (0 = hEvent) Then
      Exit Sub
   End If

   ' Signal the event.
   bRet = SetEvent(hEvent)

#If DEBUGBUILD Then
   If (0 = bRet) Then
      MsgBox ("SignalManualResetEvent failed to set event")
   End If
#End If

   ' Close the handle.
   bRet = CloseHandle(hEvent)

#If DEBUGBUILD Then
   If (0 = bRet) Then
      MsgBox ("SignalManualResetEvent CloseHandle failed!")
   End If
#End If

End Sub

